@@ -1,7 +1,7 @@
 /* Extract files from a tar archive.
 
-   Copyright 1988, 1992, 1993, 1994, 1996, 1997, 1998, 1999, 2000,
-   2001 Free Software Foundation, Inc.
+   Copyright (C) 1988, 1992, 1993, 1994, 1996, 1997, 1998, 1999, 2000,
+   2001, 2003 Free Software Foundation, Inc.
 
    Written by John Gilmore, on 1985-11-19.
 
@@ -21,6 +21,7 @@
 
 #include "system.h"
 #include <quotearg.h>
+#include <errno.h>
 
 #if HAVE_UTIME_H
 # include <utime.h>
@@ -34,7 +35,7 @@ struct utimbuf
 
 #include "common.h"
 
-int we_are_root;		/* true if our effective uid == 0 */
+bool we_are_root;		/* true if our effective uid == 0 */
 static mode_t newdir_umask;	/* umask when creating new directories */
 static mode_t current_umask;	/* current umask (which is set to 0 if -p) */
 
@@ -129,14 +130,15 @@ extr_init (void)
 }
 
 /* If restoring permissions, restore the mode for FILE_NAME from
-   information given in *STAT_INFO (where *CURRENT_STAT_INFO gives
-   the current status if CURRENT_STAT_INFO is nonzero); otherwise invert the
+   information given in *STAT_INFO (where *CUR_INFO gives
+   the current status if CUR_INFO is nonzero); otherwise invert the
    INVERT_PERMISSIONS bits from the file's current permissions.
    PERMSTATUS specifies the status of the file's permissions.
    TYPEFLAG specifies the type of the file.  */
 static void
-set_mode (char const *file_name, struct stat const *stat_info,
-	  struct stat const *current_stat_info,
+set_mode (char const *file_name,
+	  struct stat const *stat_info,
+	  struct stat const *cur_info,
 	  mode_t invert_permissions, enum permstatus permstatus,
 	  char typeflag)
 {
@@ -168,16 +170,16 @@ set_mode (char const *file_name, struct stat const *stat_info,
 	 that we created, so there's no point optimizing this code for
 	 other cases.  */
       struct stat st;
-      if (! current_stat_info)
+      if (! cur_info)
 	{
 	  if (stat (file_name, &st) != 0)
 	    {
 	      stat_error (file_name);
 	      return;
 	    }
-	  current_stat_info = &st;
+	  cur_info = &st;
 	}
-      mode = current_stat_info->st_mode ^ invert_permissions;
+      mode = cur_info->st_mode ^ invert_permissions;
     }
 
   if (chmod (file_name, mode) != 0)
@@ -189,14 +191,17 @@ static void
 check_time (char const *file_name, time_t t)
 {
   time_t now;
-  if (start_time < t && (now = time (0)) < t)
+  if (t <= 0)
+    WARN ((0, 0, _("%s: implausibly old time stamp %s"),
+	   file_name, tartime (t)));
+  else if (start_time < t && (now = time (0)) < t)
     WARN ((0, 0, _("%s: time stamp %s is %lu s in the future"),
 	   file_name, tartime (t), (unsigned long) (t - now)));
 }
 
 /* Restore stat attributes (owner, group, mode and times) for
    FILE_NAME, using information given in *STAT_INFO.
-   If CURRENT_STAT_INFO is nonzero, *CURRENT_STAT_INFO is the
+   If CUR_INFO is nonzero, *CUR_INFO is the
    file's currernt status.
    If not restoring permissions, invert the
    INVERT_PERMISSIONS bits from the file's current permissions.
@@ -209,8 +214,9 @@ check_time (char const *file_name, time_t t)
    punt for the rest.  Sigh!  */
 
 static void
-set_stat (char const *file_name, struct stat const *stat_info,
-	  struct stat const *current_stat_info,
+set_stat (char const *file_name,
+	  struct stat const *stat_info,
+	  struct stat const *cur_info,
 	  mode_t invert_permissions, enum permstatus permstatus,
 	  char typeflag)
 {
@@ -240,8 +246,8 @@ set_stat (char const *file_name, struct stat const *stat_info,
 	    utime_error (file_name);
 	  else
 	    {
-	      check_time (file_name, stat_info->st_atime);
-	      check_time (file_name, stat_info->st_mtime);
+	      check_time (file_name, utimbuf.actime);
+	      check_time (file_name, utimbuf.modtime);
 	    }
 	}
 
@@ -249,7 +255,7 @@ set_stat (char const *file_name, struct stat const *stat_info,
 	 done, it is not possible anymore to change file permissions, so we
 	 have to set permissions prior to possibly giving files away.  */
 
-      set_mode (file_name, stat_info, current_stat_info,
+      set_mode (file_name, stat_info, cur_info,
 		invert_permissions, permstatus, typeflag);
     }
 
@@ -329,9 +335,9 @@ repair_delayed_set_stat (char const *dir_name,
       if (st.st_dev == dir_stat_info->st_dev
 	  && st.st_ino == dir_stat_info->st_ino)
 	{
-	  data->stat_info = current_stat;
-	  data->invert_permissions = (MODE_RWX
-				      & (current_stat.st_mode ^ st.st_mode));
+	  data->stat_info = current_stat_info.stat;
+	  data->invert_permissions =
+	    (MODE_RWX & (current_stat_info.stat.st_mode ^ st.st_mode));
 	  data->permstatus = ARCHIVED_PERMSTATUS;
 	  return;
 	}
@@ -385,7 +391,7 @@ make_directories (char *file_name)
 	     invert_permissions is zero, because
 	     repair_delayed_set_stat may need to update the struct.  */
 	  delay_set_stat (file_name,
-			  &current_stat /* ignored */,
+			  &current_stat_info.stat /* ignored */,
 			  invert_permissions, INTERDIR_PERMSTATUS);
 
 	  print_for_mkdir (file_name, cursor - file_name, mode);
@@ -397,13 +403,13 @@ make_directories (char *file_name)
 
       *cursor = '/';
 
-      if (errno == EEXIST
-#if MSDOS
-	  /* Turbo C mkdir gives a funny errno.  */
-	  || errno == EACCES
-#endif
-	  )
-	/* Directory already exists.  */
+      if (errno == EEXIST)
+	continue;	        /* Directory already exists.  */
+      else if ((errno == ENOSYS /* Automounted dirs on Solaris return
+				   this. Reported by Warren Hyde
+				   <Warren.Hyde@motorola.com> */
+	       || ERRNO_IS_EACCES)  /* Turbo C mkdir gives a funny errno.  */
+	       && access (file_name, W_OK) == 0)
 	continue;
 
       /* Some other error in the mkdir.  We return to the caller.  */
@@ -417,7 +423,7 @@ make_directories (char *file_name)
    Return zero if extraction should not proceed.  */
 
 static int
-prepare_to_extract (char const *file_name)
+prepare_to_extract (char const *file_name, bool directory)
 {
   if (to_stdout_option)
     return 0;
@@ -454,7 +460,7 @@ maybe_recoverable (char *file_name, int *interdir_made)
 	  return 0;
 
 	case DEFAULT_OLD_FILES:
-	case OVERWRITE_OLD_DIRS:
+	case NO_OVERWRITE_DIR_OLD_FILES:
 	case OVERWRITE_OLD_FILES:
 	  {
 	    int r = remove_any_file (file_name, 0);
@@ -480,61 +486,6 @@ maybe_recoverable (char *file_name, int *interdir_made)
     }
 }
 
-static void
-extract_sparse_file (int fd, off_t *sizeleft, off_t totalsize, char *name)
-{
-  int sparse_ind = 0;
-
-  /* assuming sizeleft is initially totalsize */
-
-  while (*sizeleft > 0)
-    {
-      size_t written;
-      size_t count;
-      union block *data_block = find_next_block ();
-      if (! data_block)
-	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
-	  return;
-	}
-      if (lseek (fd, sparsearray[sparse_ind].offset, SEEK_SET) < 0)
-	{
-	  seek_error_details (name, sparsearray[sparse_ind].offset);
-	  return;
-	}
-      written = sparsearray[sparse_ind++].numbytes;
-      while (written > BLOCKSIZE)
-	{
-	  count = full_write (fd, data_block->buffer, BLOCKSIZE);
-	  written -= count;
-	  *sizeleft -= count;
-	  if (count != BLOCKSIZE)
-	    {
-	      write_error_details (name, count, BLOCKSIZE);
-	      return;
-	    }
-	  set_next_block_after (data_block);
-	  data_block = find_next_block ();
-	  if (! data_block)
-	    {
-	      ERROR ((0, 0, _("Unexpected EOF in archive")));
-	      return;
-	    }
-	}
-
-      count = full_write (fd, data_block->buffer, written);
-      *sizeleft -= count;
-
-      if (count != written)
-	{
-	  write_error_details (name, count, written);
-	  return;
-	}
-
-      set_next_block_after (data_block);
-    }
-}
-
 /* Fix the statuses of all directories whose statuses need fixing, and
    which are not ancestors of FILE_NAME.  If AFTER_SYMLINKS is
    nonzero, do this for all such directories; otherwise, stop at the
@@ -551,7 +502,7 @@ apply_nonancestor_delayed_set_stat (char const *file_name, bool after_symlinks)
       struct delayed_set_stat *data = delayed_set_stat_head;
       bool skip_this_one = 0;
       struct stat st;
-      struct stat const *current_stat_info = 0;
+      struct stat const *cur_info = 0;
 
       check_for_renamed_directories |= data->after_symlinks;
 
@@ -565,7 +516,7 @@ apply_nonancestor_delayed_set_stat (char const *file_name, bool after_symlinks)
 
       if (check_for_renamed_directories)
 	{
-	  current_stat_info = &st;
+	  cur_info = &st;
 	  if (stat (data->file_name, &st) != 0)
 	    {
 	      stat_error (data->file_name);
@@ -582,7 +533,7 @@ apply_nonancestor_delayed_set_stat (char const *file_name, bool after_symlinks)
 	}
 
       if (! skip_this_one)
-	set_stat (data->file_name, &data->stat_info, current_stat_info,
+	set_stat (data->file_name, &data->stat_info, cur_info,
 		  data->invert_permissions, data->permstatus, DIRTYPE);
 
       delayed_set_stat_head = data->next;
@@ -598,23 +549,19 @@ extract_archive (void)
   int fd;
   int status;
   size_t count;
-  size_t name_length;
   size_t written;
   int openflag;
   mode_t mode;
   off_t size;
-  size_t skipcrud;
-  int counter;
+  off_t file_size;
   int interdir_made = 0;
   char typeflag;
-  union block *exhdr;
-
-#define CURRENT_FILE_NAME (skipcrud + current_file_name)
+  char *file_name;
 
   set_next_block_after (current_header);
-  decode_header (current_header, &current_stat, &current_format, 1);
+  decode_header (current_header, &current_stat_info, &current_format, 1);
 
-  if (interactive_option && !confirm ("extract", current_file_name))
+  if (interactive_option && !confirm ("extract", current_stat_info.file_name))
     {
       skip_member ();
       return;
@@ -623,48 +570,30 @@ extract_archive (void)
   /* Print the block from current_header and current_stat.  */
 
   if (verbose_option)
-    print_header ();
+    print_header (&current_stat_info, -1);
 
-  /* Check for fully specified file names and other atrocities.  */
-
-  skipcrud = 0;
-  if (! absolute_names_option)
+  file_name = safer_name_suffix (current_stat_info.file_name, false);
+  if (strip_path_elements)
     {
-      if (contains_dot_dot (CURRENT_FILE_NAME))
+      size_t prefix_len = stripped_prefix_len (file_name, strip_path_elements);
+      if (prefix_len == (size_t) -1)
 	{
-	  ERROR ((0, 0, _("%s: Member name contains `..'"),
-		  quotearg_colon (CURRENT_FILE_NAME)));
 	  skip_member ();
 	  return;
 	}
-
-      skipcrud = FILESYSTEM_PREFIX_LEN (current_file_name);
-      while (ISSLASH (CURRENT_FILE_NAME[0]))
-	skipcrud++;
-
-      if (skipcrud)
-	{
-	  static int warned_once;
-	  
-	  if (!warned_once)
-	    {
-	      warned_once = 1;
-	      WARN ((0, 0, _("Removing leading `%.*s' from member names"),
-		     (int) skipcrud, current_file_name));
-	    }
-	}
+      file_name += prefix_len;
     }
-
-  apply_nonancestor_delayed_set_stat (CURRENT_FILE_NAME, 0);
+  
+  apply_nonancestor_delayed_set_stat (file_name, 0);
 
   /* Take a safety backup of a previously existing file.  */
 
   if (backup_option && !to_stdout_option)
-    if (!maybe_backup_file (CURRENT_FILE_NAME, 0))
+    if (!maybe_backup_file (file_name, 0))
       {
 	int e = errno;
 	ERROR ((0, e, _("%s: Was unable to backup this file"),
-		quotearg_colon (CURRENT_FILE_NAME)));
+		quotearg_colon (file_name)));
 	skip_member ();
 	return;
       }
@@ -672,82 +601,14 @@ extract_archive (void)
   /* Extract the archive entry according to its type.  */
 
   typeflag = current_header->header.typeflag;
+  /*KLUDGE */
+  if (current_format == POSIX_FORMAT
+      && current_stat_info.archive_file_size != current_stat_info.stat.st_size)
+	  typeflag = GNUTYPE_SPARSE;
+  
   switch (typeflag)
     {
-      /* JK - What we want to do if the file is sparse is loop through
-	 the array of sparse structures in the header and read in and
-	 translate the character strings representing 1) the offset at
-	 which to write and 2) how many bytes to write into numbers,
-	 which we store into the scratch array, "sparsearray".  This
-	 array makes our life easier the same way it did in creating the
-	 tar file that had to deal with a sparse file.
-
-	 After we read in the first five (at most) sparse structures, we
-	 check to see if the file has an extended header, i.e., if more
-	 sparse structures are needed to describe the contents of the new
-	 file.  If so, we read in the extended headers and continue to
-	 store their contents into the sparsearray.  */
-
     case GNUTYPE_SPARSE:
-      sp_array_size = 10;
-      sparsearray =
-	xmalloc (sp_array_size * sizeof (struct sp_array));
-
-      for (counter = 0; counter < SPARSES_IN_OLDGNU_HEADER; counter++)
-	{
-	  struct sparse const *s = &current_header->oldgnu_header.sp[counter];
-	  sparsearray[counter].offset = OFF_FROM_HEADER (s->offset);
-	  sparsearray[counter].numbytes = SIZE_FROM_HEADER (s->numbytes);
-	  if (!sparsearray[counter].numbytes)
-	    break;
-	}
-
-      if (current_header->oldgnu_header.isextended)
-	{
-	  /* Read in the list of extended headers and translate them
-	     into the sparsearray as before.  Note that this
-	     invalidates current_header.  */
-
-	  /* static */ int ind = SPARSES_IN_OLDGNU_HEADER;
-
-	  while (1)
-	    {
-	      exhdr = find_next_block ();
-	      if (! exhdr)
-		{
-		  ERROR ((0, 0, _("Unexpected EOF in archive")));
-		  return;
-		}
-	      for (counter = 0; counter < SPARSES_IN_SPARSE_HEADER; counter++)
-		{
-		  struct sparse const *s = &exhdr->sparse_header.sp[counter];
-		  if (counter + ind > sp_array_size - 1)
-		    {
-		      /* Realloc the scratch area since we've run out of
-			 room.  */
-
-		      sp_array_size *= 2;
-		      sparsearray =
-			xrealloc (sparsearray,
-				  sp_array_size * sizeof (struct sp_array));
-		    }
-		  if (s->numbytes[0] == 0)
-		    break;
-		  sparsearray[counter + ind].offset =
-		    OFF_FROM_HEADER (s->offset);
-		  sparsearray[counter + ind].numbytes =
-		    SIZE_FROM_HEADER (s->numbytes);
-		}
-	      if (!exhdr->sparse_header.isextended)
-		break;
-	      else
-		{
-		  ind += SPARSES_IN_SPARSE_HEADER;
-		  set_next_block_after (exhdr);
-		}
-	    }
-	  set_next_block_after (exhdr);
-	}
       /* Fall through.  */
 
     case AREGTYPE:
@@ -757,9 +618,7 @@ extract_archive (void)
       /* Appears to be a file.  But BSD tar uses the convention that a slash
 	 suffix means a directory.  */
 
-      name_length = strlen (CURRENT_FILE_NAME);
-      if (FILESYSTEM_PREFIX_LEN (CURRENT_FILE_NAME) < name_length
-	  && CURRENT_FILE_NAME[name_length - 1] == '/')
+      if (current_stat_info.had_trailing_slash)
 	goto really_dir;
 
       /* FIXME: deal with protection issues.  */
@@ -769,7 +628,7 @@ extract_archive (void)
 		  | (old_files_option == OVERWRITE_OLD_FILES
 		     ? O_TRUNC
 		     : O_EXCL));
-      mode = current_stat.st_mode & MODE_RWX & ~ current_umask;
+      mode = current_stat_info.stat.st_mode & MODE_RWX & ~ current_umask;
 
       if (to_stdout_option)
 	{
@@ -777,7 +636,7 @@ extract_archive (void)
 	  goto extract_file;
 	}
 
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
+      if (! prepare_to_extract (file_name, 0))
 	{
 	  skip_member ();
 	  if (backup_option)
@@ -790,10 +649,9 @@ extract_archive (void)
 	 the open call that creates them.  */
 
       if (typeflag == CONTTYPE)
-	fd = open (CURRENT_FILE_NAME, openflag | O_CTG,
-		   mode, current_stat.st_size);
+	fd = open (file_name, openflag | O_CTG, mode, current_stat_info.stat.st_size);
       else
-	fd = open (CURRENT_FILE_NAME, openflag, mode);
+	fd = open (file_name, openflag, mode);
 
 #else /* not O_CTG */
       if (typeflag == CONTTYPE)
@@ -806,16 +664,16 @@ extract_archive (void)
 	      WARN ((0, 0, _("Extracting contiguous files as regular files")));
 	    }
 	}
-      fd = open (CURRENT_FILE_NAME, openflag, mode);
+      fd = open (file_name, openflag, mode);
 
 #endif /* not O_CTG */
 
       if (fd < 0)
 	{
-	  if (maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	  if (maybe_recoverable (file_name, &interdir_made))
 	    goto again_file;
 
-	  open_error (CURRENT_FILE_NAME);
+	  open_error (file_name);
 	  skip_member ();
 	  if (backup_option)
 	    undo_last_backup ();
@@ -825,29 +683,15 @@ extract_archive (void)
     extract_file:
       if (typeflag == GNUTYPE_SPARSE)
 	{
-	  char *name;
-	  size_t name_length_bis;
-
-	  /* Kludge alert.  NAME is assigned to header.name because
-	     during the extraction, the space that contains the header
-	     will get scribbled on, and the name will get munged, so any
-	     error messages that happen to contain the filename will look
-	     REAL interesting unless we do this.  */
-
-	  name_length_bis = strlen (CURRENT_FILE_NAME) + 1;
-	  name = xmalloc (name_length_bis);
-	  memcpy (name, CURRENT_FILE_NAME, name_length_bis);
-	  size = current_stat.st_size;
-	  extract_sparse_file (fd, &size, current_stat.st_size, name);
-	  free (sparsearray);
+	  sparse_extract_file (fd, &current_stat_info, &size);
 	}
       else
-	for (size = current_stat.st_size; size > 0; )
+	for (size = current_stat_info.stat.st_size; size > 0; )
 	  {
 	    if (multi_volume_option)
 	      {
-		assign_string (&save_name, current_file_name);
-		save_totsize = current_stat.st_size;
+		assign_string (&save_name, current_stat_info.file_name);
+		save_totsize = current_stat_info.stat.st_size;
 		save_sizeleft = size;
 	      }
 
@@ -874,7 +718,7 @@ extract_archive (void)
 				  (data_block->buffer + written - 1));
 	    if (count != written)
 	      {
-		write_error_details (CURRENT_FILE_NAME, count, written);
+		write_error_details (file_name, count, written);
 		break;
 	      }
 	  }
@@ -893,12 +737,12 @@ extract_archive (void)
       status = close (fd);
       if (status < 0)
 	{
-	  close_error (CURRENT_FILE_NAME);
+	  close_error (file_name);
 	  if (backup_option)
 	    undo_last_backup ();
 	}
 
-      set_stat (CURRENT_FILE_NAME, &current_stat, 0, 0,
+      set_stat (file_name, &current_stat_info.stat, 0, 0,
 		(old_files_option == OVERWRITE_OLD_FILES
 		 ? UNKNOWN_PERMSTATUS
 		 : ARCHIVED_PERMSTATUS),
@@ -907,23 +751,23 @@ extract_archive (void)
 
     case SYMTYPE:
 #ifdef HAVE_SYMLINK
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
+      if (! prepare_to_extract (file_name, 0))
 	break;
 
       if (absolute_names_option
-	  || ! (ISSLASH (current_link_name
-			 [FILESYSTEM_PREFIX_LEN (current_link_name)])
-		|| contains_dot_dot (current_link_name)))
+	  || ! (ISSLASH (current_stat_info.link_name
+			 [FILESYSTEM_PREFIX_LEN (current_stat_info.link_name)])
+		|| contains_dot_dot (current_stat_info.link_name)))
 	{
-	  while (status = symlink (current_link_name, CURRENT_FILE_NAME),
+	  while (status = symlink (current_stat_info.link_name, file_name),
 		 status != 0)
-	    if (!maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	    if (!maybe_recoverable (file_name, &interdir_made))
 	      break;
 
 	  if (status == 0)
-	    set_stat (CURRENT_FILE_NAME, &current_stat, 0, 0, 0, SYMTYPE);
+	    set_stat (file_name, &current_stat_info.stat, 0, 0, 0, SYMTYPE);
 	  else
-	    symlink_error (current_link_name, CURRENT_FILE_NAME);
+	    symlink_error (current_stat_info.link_name, file_name);
 	}
       else
 	{
@@ -932,46 +776,46 @@ extract_archive (void)
 	     will be replaced after other extraction is done.  */
 	  struct stat st;
 
-	  while (fd = open (CURRENT_FILE_NAME, O_WRONLY | O_CREAT | O_EXCL, 0),
+	  while (fd = open (file_name, O_WRONLY | O_CREAT | O_EXCL, 0),
 		 fd < 0)
-	    if (! maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	    if (! maybe_recoverable (file_name, &interdir_made))
 	      break;
 
 	  status = -1;
 	  if (fd < 0)
-	    open_error (CURRENT_FILE_NAME);
+	    open_error (file_name);
 	  else if (fstat (fd, &st) != 0)
 	    {
-	      stat_error (CURRENT_FILE_NAME);
+	      stat_error (file_name);
 	      close (fd);
 	    }
 	  else if (close (fd) != 0)
-	    close_error (CURRENT_FILE_NAME);
+	    close_error (file_name);
 	  else
 	    {
 	      struct delayed_set_stat *h;
 	      struct delayed_symlink *p =
 		xmalloc (offsetof (struct delayed_symlink, target)
-			 + strlen (current_link_name) + 1);
+			 + strlen (current_stat_info.link_name) + 1);
 	      p->next = delayed_symlink_head;
 	      delayed_symlink_head = p;
 	      p->dev = st.st_dev;
 	      p->ino = st.st_ino;
 	      p->mtime = st.st_mtime;
-	      p->uid = current_stat.st_uid;
-	      p->gid = current_stat.st_gid;
+	      p->uid = current_stat_info.stat.st_uid;
+	      p->gid = current_stat_info.stat.st_gid;
 	      p->sources = xmalloc (offsetof (struct string_list, string)
-				    + strlen (CURRENT_FILE_NAME) + 1);
+				    + strlen (file_name) + 1);
 	      p->sources->next = 0;
-	      strcpy (p->sources->string, CURRENT_FILE_NAME);
-	      strcpy (p->target, current_link_name);
+	      strcpy (p->sources->string, file_name);
+	      strcpy (p->target, current_stat_info.link_name);
 
 	      h = delayed_set_stat_head;
 	      if (h && ! h->after_symlinks
-		  && strncmp (CURRENT_FILE_NAME, h->file_name, h->file_name_len) == 0
-		  && ISSLASH (CURRENT_FILE_NAME[h->file_name_len])
-		  && (base_name (CURRENT_FILE_NAME)
-		      == CURRENT_FILE_NAME + h->file_name_len + 1))
+		  && strncmp (file_name, h->file_name, h->file_name_len) == 0
+		  && ISSLASH (file_name[h->file_name_len])
+		  && (base_name (file_name)
+		      == file_name + h->file_name_len + 1))
 		{
 		  do
 		    {
@@ -1012,22 +856,24 @@ extract_archive (void)
 #endif
 
     case LNKTYPE:
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
+      if (! prepare_to_extract (file_name, 0))
 	break;
 
     again_link:
       {
+	char const *link_name = safer_name_suffix (current_stat_info.link_name,
+	                                           true);
 	struct stat st1, st2;
 	int e;
 
 	/* MSDOS does not implement links.  However, djgpp's link() actually
 	   copies the file.  */
-	status = link (current_link_name, CURRENT_FILE_NAME);
+	status = link (link_name, file_name);
 
 	if (status == 0)
 	  {
 	    struct delayed_symlink *ds = delayed_symlink_head;
-	    if (ds && stat (current_link_name, &st1) == 0)
+	    if (ds && stat (link_name, &st1) == 0)
 	      for (; ds; ds = ds->next)
 		if (ds->dev == st1.st_dev
 		    && ds->ino == st1.st_ino
@@ -1035,27 +881,27 @@ extract_archive (void)
 		  {
 		    struct string_list *p = 
 		      xmalloc (offsetof (struct string_list, string)
-			       + strlen (CURRENT_FILE_NAME) + 1);
-		    strcpy (p->string, CURRENT_FILE_NAME);
+			       + strlen (file_name) + 1);
+		    strcpy (p->string, file_name);
 		    p->next = ds->sources;
 		    ds->sources = p;
 		    break;
 		  }
 	    break;
 	  }
-	if (maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	if (maybe_recoverable (file_name, &interdir_made))
 	  goto again_link;
 
 	if (incremental_option && errno == EEXIST)
 	  break;
 	e = errno;
-	if (stat (current_link_name, &st1) == 0
-	    && stat (CURRENT_FILE_NAME, &st2) == 0
+	if (stat (link_name, &st1) == 0
+	    && stat (file_name, &st2) == 0
 	    && st1.st_dev == st2.st_dev
 	    && st1.st_ino == st2.st_ino)
 	  break;
 
-	link_error (current_link_name, CURRENT_FILE_NAME);
+	link_error (link_name, file_name);
 	if (backup_option)
 	  undo_last_backup ();
       }
@@ -1063,52 +909,52 @@ extract_archive (void)
 
 #if S_IFCHR
     case CHRTYPE:
-      current_stat.st_mode |= S_IFCHR;
+      current_stat_info.stat.st_mode |= S_IFCHR;
       goto make_node;
 #endif
 
 #if S_IFBLK
     case BLKTYPE:
-      current_stat.st_mode |= S_IFBLK;
+      current_stat_info.stat.st_mode |= S_IFBLK;
 #endif
 
 #if S_IFCHR || S_IFBLK
     make_node:
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
+      if (! prepare_to_extract (file_name, 0))
 	break;
 
-      status = mknod (CURRENT_FILE_NAME, current_stat.st_mode,
-		      current_stat.st_rdev);
+      status = mknod (file_name, current_stat_info.stat.st_mode,
+		      current_stat_info.stat.st_rdev);
       if (status != 0)
 	{
-	  if (maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	  if (maybe_recoverable (file_name, &interdir_made))
 	    goto make_node;
-	  mknod_error (CURRENT_FILE_NAME);
+	  mknod_error (file_name);
 	  if (backup_option)
 	    undo_last_backup ();
 	  break;
 	};
-      set_stat (CURRENT_FILE_NAME, &current_stat, 0, 0,
+      set_stat (file_name, &current_stat_info.stat, 0, 0,
 		ARCHIVED_PERMSTATUS, typeflag);
       break;
 #endif
 
 #if HAVE_MKFIFO || defined mkfifo
     case FIFOTYPE:
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
+      if (! prepare_to_extract (file_name, 0))
 	break;
 
-      while (status = mkfifo (CURRENT_FILE_NAME, current_stat.st_mode),
+      while (status = mkfifo (file_name, current_stat_info.stat.st_mode),
 	     status != 0)
-	if (!maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	if (!maybe_recoverable (file_name, &interdir_made))
 	  break;
 
       if (status == 0)
-	set_stat (CURRENT_FILE_NAME, &current_stat, 0, 0,
+	set_stat (file_name, &current_stat_info.stat, NULL, 0,
 		  ARCHIVED_PERMSTATUS, typeflag);
       else
 	{
-	  mkfifo_error (CURRENT_FILE_NAME);
+	  mkfifo_error (file_name);
 	  if (backup_option)
 	    undo_last_backup ();
 	}
@@ -1117,48 +963,43 @@ extract_archive (void)
 
     case DIRTYPE:
     case GNUTYPE_DUMPDIR:
-      name_length = strlen (CURRENT_FILE_NAME);
-
     really_dir:
-      /* Remove any redundant trailing "/"s.  */
-      while (FILESYSTEM_PREFIX_LEN (CURRENT_FILE_NAME) < name_length
-	     && CURRENT_FILE_NAME[name_length - 1] == '/')
-	name_length--;
-      CURRENT_FILE_NAME[name_length] = '\0';
-
       if (incremental_option)
 	{
 	  /* Read the entry and delete files that aren't listed in the
 	     archive.  */
 
-	  gnu_restore (skipcrud);
+	  gnu_restore (file_name);
 	}
       else if (typeflag == GNUTYPE_DUMPDIR)
 	skip_member ();
 
-      if (! prepare_to_extract (CURRENT_FILE_NAME))
-	break;
-
-      mode = ((current_stat.st_mode
+      mode = ((current_stat_info.stat.st_mode
 	       | (we_are_root ? 0 : MODE_WXUSR))
 	      & MODE_RWX);
 
+      status = prepare_to_extract (file_name, 1);
+      if (status == 0)
+	break;
+      if (status < 0)
+	goto directory_exists;
+
     again_dir:
-      status = mkdir (CURRENT_FILE_NAME, mode);
+      status = mkdir (file_name, mode);
 
       if (status != 0)
 	{
 	  if (errno == EEXIST
 	      && (interdir_made
-		  || old_files_option == OVERWRITE_OLD_DIRS
+		  || old_files_option == DEFAULT_OLD_FILES
 		  || old_files_option == OVERWRITE_OLD_FILES))
 	    {
 	      struct stat st;
-	      if (stat (CURRENT_FILE_NAME, &st) == 0)
+	      if (stat (file_name, &st) == 0)
 		{
 		  if (interdir_made)
 		    {
-		      repair_delayed_set_stat (CURRENT_FILE_NAME, &st);
+		      repair_delayed_set_stat (file_name, &st);
 		      break;
 		    }
 		  if (S_ISDIR (st.st_mode))
@@ -1170,12 +1011,12 @@ extract_archive (void)
 	      errno = EEXIST;
 	    }
 	
-	  if (maybe_recoverable (CURRENT_FILE_NAME, &interdir_made))
+	  if (maybe_recoverable (file_name, &interdir_made))
 	    goto again_dir;
 
 	  if (errno != EEXIST)
 	    {
-	      mkdir_error (CURRENT_FILE_NAME);
+	      mkdir_error (file_name);
 	      if (backup_option)
 		undo_last_backup ();
 	      break;
@@ -1184,10 +1025,10 @@ extract_archive (void)
 
     directory_exists:
       if (status == 0
-	  || old_files_option == OVERWRITE_OLD_DIRS
+	  || old_files_option == DEFAULT_OLD_FILES
 	  || old_files_option == OVERWRITE_OLD_FILES)
-	delay_set_stat (CURRENT_FILE_NAME, &current_stat,
-			MODE_RWX & (mode ^ current_stat.st_mode),
+	delay_set_stat (file_name, &current_stat_info.stat,
+			MODE_RWX & (mode ^ current_stat_info.stat.st_mode),
 			(status == 0
 			 ? ARCHIVED_PERMSTATUS
 			 : UNKNOWN_PERMSTATUS));
@@ -1195,7 +1036,7 @@ extract_archive (void)
 
     case GNUTYPE_VOLHDR:
       if (verbose_option)
-	fprintf (stdlis, _("Reading %s\n"), quote (current_file_name));
+	fprintf (stdlis, _("Reading %s\n"), quote (current_stat_info.file_name));
       break;
 
     case GNUTYPE_NAMES:
@@ -1205,7 +1046,7 @@ extract_archive (void)
     case GNUTYPE_MULTIVOL:
       ERROR ((0, 0,
 	      _("%s: Cannot extract -- file is continued from another volume"),
-	      quotearg_colon (current_file_name)));
+	      quotearg_colon (current_stat_info.file_name)));
       skip_member ();
       if (backup_option)
 	undo_last_backup ();
@@ -1222,11 +1063,9 @@ extract_archive (void)
     default:
       WARN ((0, 0,
 	     _("%s: Unknown file type '%c', extracted as normal file"),
-	     quotearg_colon (CURRENT_FILE_NAME), typeflag));
+	     quotearg_colon (file_name), typeflag));
       goto again_file;
     }
-
-#undef CURRENT_FILE_NAME
 }
 
 /* Extract the symbolic links whose final extraction were delayed.  */
